local MyName="vZifnabsGatherer"
local Me=CreateFrame("Frame",MyName)
local mypfx="vZifnabs_Gatherer"
vZifnabs.mod.gatherer=Me
local mod=vZifnabs.mod
local db={}
local L={}
local playername=UnitName("player")

local fishing={}
local current_zone=nil

local mmlist={} -- {mmlist[n]={x=x,y=y,gathertype=gathertype,id=itemid,tt=tooltip} all items to be shown in current zone
local mmshow={n=0,mmn=0,chunk=0} -- {n=next_to_check,mmn=next_widget_to_use}
local MAXMM=20
local MAXWM=50
local widget={mm={},wm={},config={}}
local TYPE={HERB="herb",ORE="ore",TREASURE="treasure",FISH="fish",CRATE="crate",SPECIAL="special"}
local id2item={}
local config={}

-- -----------------------------------------------------------------
-- INITIALIZE
-- -----------------------------------------------------------------
function Me:Initialize()
	fishing=nil
	db,L=mod.core:InitStart("gatherer")
	if not db.gatherer then db.gatherer={nodedistance=2,mindistance=20,maxdistance=100,mmsize=15,wmsize=15,show={},node={}} end
	if not db.gatherer.show[playername] then db.gatherer.show[playername]={[TYPE.TREASURE]=true,[TYPE.HERB]=true,[TYPE.ORE]=true,[TYPE.FISH]=true,[TYPE.CRATE]=true,[TYPE.SPECIAL]=true} end

	mod.core:RegisterEvent("ONUPDATE","gatherer",Me.OnUpdate)
	mod.core:RegisterEvent("SPELLCAST_START","gatherer",Me.SPELLCAST_START)
	mod.core:RegisterEvent("SPELLCAST_CHANNEL_START","gatherer",Me.SPELLCAST_CHANNEL_START)
	mod.core:RegisterEvent("SPELLCAST_CHANNEL_STOP","gatherer",Me.SPELLCAST_CHANNEL_STOP)
	mod.core:RegisterEvent("WORLD_MAP_UPDATE","gatherer",Me.WORLD_MAP_UPDATE);
	
	for n=1,MAXMM,1 do widget.mm[n]=getglobal(mypfx.."_MM_"..n) end
	for n=1,MAXWM,1 do widget.wm[n]=getglobal(mypfx.."_WM_"..n) end
	
	Me:RedefineItems()
	
	Me:ConfigInitialize()
	mod.core:InitStop("gatherer")
end

-- -----------------------------------------------------------------
-- ON UPDATE
-- -----------------------------------------------------------------
function Me:OnUpdate(elapsed)
	if fishing then
		local cursorname=getglobal("GameTooltipTextLeft1"):GetText() or ""
		if cursorname~=fishing[fishing.n] and cursorname~="" then
			fishing.n=fishing.n+1
			fishing[fishing.n]=cursorname
		end
	end

	local unscaledx, unscaledy, mmscalex,mmscaley=mod.locator:GetMinimapScale()
	local mmradius=Minimap:GetWidth() / 2
	local wx,wy,zone,pos,zone_nr=mod.locator:GetLoc()
	if zone_nr~=current_zone then Me:ZONE_CHANGED(zone_nr) end
	
	for loop=1,mmshow.chunk,1 do
		local n=mmshow.n+1
		mmshow.n=n
		if n>mmshow.last then
			for n=mmshow.mmn+1,MAXMM,1 do widget.mm[n]:Hide() end
			mmshow.n=0
			mmshow.mmn=0
		else
			local mmx=(mmlist[n].x-wx)*mmradius/mmscalex
			local mmy=(wy-mmlist[n].y)*mmradius/mmscaley
			local mmdistance=sqrt(mmx * mmx + mmy * mmy)
			local evershow=(mmlist[n].gathertype==TYPE.TREASURE or mmlist[n].gathertype==TYPE.FISH)
			if (evershow and mmdistance<db.gatherer.maxdistance) or (mmdistance>db.gatherer.mindistance and mmdistance<db.gatherer.maxdistance) then
				if mmdistance > mmradius-4 then
					mmx=mmx*(mmradius-4)/mmdistance
					mmy=mmy*(mmradius-4)/mmdistance
				end
				local mmn=mmshow.mmn+1
				mmshow.mmn=mmn
				if mmn<=MAXMM then
					widget.mm[mmn]:SetPoint("CENTER","Minimap","CENTER",mmx,mmy)
					Me:ShowIcon(widget.mm[mmn],mmlist[n].gathertype,mmlist[n].id,mmlist[n].tt.."\n>> "..math.floor(10*mmdistance)/10)
					local alpha=1
					if mmdistance<db.gatherer.mindistance+(db.gatherer.maxdistance-db.gatherer.mindistance)/2 then alpha=0.5 end
					if evershow then 
						if mmdistance>db.gatherer.mindistance then alpha=1 else alpha=0.5 end
					end
					widget.mm[mmn]:SetAlpha(alpha)
				end
			end
		end
	end
end

-- -----------------------------------------------------------------
-- MAP CHANGED
-- -----------------------------------------------------------------
function Me:ZONE_CHANGED(zone_nr)
	mmlist={}
	if db.gatherer.node[zone_nr] then
		for gathertype,list in pairs(db.gatherer.node[zone_nr]) do
			if db.gatherer.show[playername][gathertype] then
				for n=1,list.n,1 do
					local x,y,maxid,tooltip=Me:GetNodeData(list[n])
					if maxid~=nil then Me:add2list(x,y,gathertype,maxid,tooltip) end
				end
			end
		end
	end
	current_zone=zone_nr
	mmshow={n=0,mmn=0,last=getn(mmlist),chunk=math.max(2,math.ceil(getn(mmlist)/4))}
end

-- -----------------------------------------------------------------
-- WORLD MAP CHANGED
-- -----------------------------------------------------------------
function Me:WORLD_MAP_UPDATE()
	if WorldMapFrame:IsVisible() then
		local continent = GetCurrentMapContinent()
		local zonelist={GetMapZones(continent)}
		local zone=zonelist[GetCurrentMapZone()]
		if L.locator.zonereplace[zone] then zone=L.locator.zonereplace[zone] end
		local zone_nr=mod.locator:GetZoneNr(zone)
		local wmn=0
		for gathertype,list in pairs(db.gatherer.node[zone_nr] or {}) do
			if db.gatherer.show[playername][gathertype] then 
				for n=1,list.n,1 do
					local x,y,maxid,tooltip=Me:GetNodeData(list[n])
					if maxid~=nil then
						local wx=x*WorldMapDetailFrame:GetWidth()/100
						local wy=y*WorldMapDetailFrame:GetHeight()/100
						wmn=wmn+1
						if wmn<=MAXWM then
							widget.wm[wmn]:SetPoint("CENTER","WorldMapDetailFrame","TOPLEFT",wx,-wy)
							Me:ShowIcon(widget.wm[wmn],gathertype,maxid,tooltip)
						end
					end
				end
			end
		end
		for hide=wmn+1,MAXWM,1 do widget.wm[hide]:Hide() end
	end
end

-- -----------------------------------------------------------------
-- SPELLCAST START
-- -----------------------------------------------------------------
function Me:SPELLCAST_START(event,spell,number,more)
	local cursorname=getglobal("GameTooltipTextLeft1"):GetText() or ""
	if spell==L.keyword.gatherHerb then Me:StoreLocation(TYPE.HERB,cursorname)
	elseif spell==L.keyword.mining then Me:StoreLocation(TYPE.ORE,cursorname)
	elseif spell==L.keyword.opening then Me:StoreLocation(TYPE.TREASURE,cursorname)
	elseif spell==L.keyword.picklock then Me:StoreLocation(TYPE.TREASURE,cursorname)
	end
end

-- -----------------------------------------------------------------
-- SPELLCAST CHANNEL START
-- -----------------------------------------------------------------
function Me:SPELLCAST_CHANNEL_START(event,duration,spell,rank)
	if spell==L.keyword.fishing then fishing={n=0} end
end

-- -----------------------------------------------------------------
-- SPELLCAST CHANNEL STOP
-- -----------------------------------------------------------------
function Me:SPELLCAST_CHANNEL_STOP(event,wtf,spell,rank)
	if fishing then
		local item=nil
		local unknown=nil
		for n=fishing.n,1,-1 do
			if item==nil and fishing[n]~=L.keyword.FishingBobber then 
				if L.gatherer.types[TYPE.FISH][fishing[n]] then item=fishing[n] 
				elseif unknown==nil then unknown=fishing[n] end
			end
		end
		if item==nil then item=unknown end
		if item then Me:StoreLocation(TYPE.FISH,item) end
	end
	fishing=nil
end

-- -----------------------------------------------------------------
-- STORE LOCATION
-- -----------------------------------------------------------------
function Me:StoreLocation(gathertype,item)
	local id=nil
	for ckfree=9999,9000,-1 do
		if not id2item[ckfree] then 
			id=ckfree 
			break
		end
	end
	if not L.gatherer.types[gathertype][item] then 
		mod.core:Print("DEV: gathering unknown "..gathertype.." '"..item.."'")
		L.gatherer.types[gathertype][item]={id=id,icon=gathertype.."Unknown"}
		id2item[id]={item=item,icon=gathertype.."Unknown",gathertype=gathertype}
		if not db.gatherer.unknown then db.gatherer.unknown={} end
		db.gatherer.unknown[item]={id=id,icon=gathertype.."Unknown",gathertype=gathertype}
	else
		id=L.gatherer.types[gathertype][item].id
	end
	x,y,zone,pos,zone_nr=mod.locator:GetLoc()
	if not db.gatherer.node then db.gather.node={} end
	if not db.gatherer.node[zone_nr] then db.gatherer.node[zone_nr]={} end
	if not db.gatherer.node[zone_nr][gathertype] then db.gatherer.node[zone_nr][gathertype]={n=0} end
	local entry=db.gatherer.node[zone_nr][gathertype].n+1
	local scalex,scaley=mod.locator:GetMinimapScale()
	for n=1,db.gatherer.node[zone_nr][gathertype].n,1 do
		local difx=scalex*math.abs(db.gatherer.node[zone_nr][gathertype][n].x-x)
		local dify=scaley*math.abs(db.gatherer.node[zone_nr][gathertype][n].y-y)
		if math.sqrt(difx*difx+dify*dify)<db.gatherer.nodedistance then entry=n end			
	end
	if entry>db.gatherer.node[zone_nr][gathertype].n then 
		db.gatherer.node[zone_nr][gathertype].n=entry 
		db.gatherer.node[zone_nr][gathertype][entry]={x=x,y=y,item={}}
		mod.core:Print(L.msg.newspot)
	else
		local count=0
		for item,icount in pairs(db.gatherer.node[zone_nr][gathertype][entry].item) do
			count=count+icount
		end
		-- x/y zwischen alt und neu mitteln
		db.gatherer.node[zone_nr][gathertype][entry].x=math.floor(100*(x+count*db.gatherer.node[zone_nr][gathertype][entry].x)/(count+1)+0.5)/100
		db.gatherer.node[zone_nr][gathertype][entry].y=math.floor(100*(y+count*db.gatherer.node[zone_nr][gathertype][entry].y)/(count+1)+0.5)/100
	end
	if not db.gatherer.node[zone_nr][gathertype][entry].item[id] then db.gatherer.node[zone_nr][gathertype][entry].item[id]=0 end
	db.gatherer.node[zone_nr][gathertype][entry].item[id]=db.gatherer.node[zone_nr][gathertype][entry].item[id]+1
	Me:ZONE_CHANGED(zone_nr) -- rebuild list
end

-- -----------------------------------------------------------------
-- GET NODE DATA
-- -----------------------------------------------------------------
function Me:GetNodeData(node)
	local maxcount=0
	local maxid=nil
	local tooltip=""
	local showcrate=db.gatherer.show[playername][TYPE.CRATE]
	local showspecial=db.gatherer.show[playername][TYPE.SPECIAL]
	for id,count in pairs(node.item) do
		local ignore=(not showspecial and id>=9000) or (not showcrate and id2item[id].crate==true)
		if not ignore then
			tooltip=tooltip.."\n"..count.."x "..(id2item[id].item or "???")
			if count>maxcount then
				maxcount=count
				maxid=id
			end
		end
	end
	return node.x,node.y,maxid,string.sub(tooltip,2)
end

-- -----------------------------------------------------------------
-- ADD 2 LIST
-- -----------------------------------------------------------------
function Me:add2list(x,y,gathertype,id,tooltip)
	local n=1+getn(mmlist)
	mmlist[n]={x=x,y=y,gathertype=gathertype,id=id,tt=tooltip}
end

-- -----------------------------------------------------------------
-- SHOW ICON
-- -----------------------------------------------------------------
function Me:ShowIcon(widget,gathertype,id,tooltip)
	local icon="Interface\\AddOns\\vZifnabs\\GatherIcons\\"..id2item[id].icon
	local item=id2item[id].item
	widget:SetNormalTexture(icon)
	widget.tooltipText=tooltip
	widget:Show()
end

-- -----------------------------------------------------------------
-- SHOW/HIDE Worldmap Tooltip
-- -----------------------------------------------------------------
function Me:ToolTip(widget)
	if widget.tooltipText then
		getglobal("vZifnabs_Gatherer_Tooltip_Text"):SetText(widget.tooltipText)
		getglobal("vZifnabs_Gatherer_Tooltip"):Show()
	end
end
function Me:ToolTipHide() getglobal("vZifnabs_Gatherer_Tooltip"):Hide() end

-- -----------------------------------------------------------------
-- REDEFINE ITEMS
-- -----------------------------------------------------------------
function Me:RedefineItems()
	id2item={}
	for gathertype,list in pairs(L.gatherer.types) do
		for item,info in pairs(list) do
			if info.id>=9000 then L.gatherer.types[gathertype][item]=nil end
		end
	end
	if db.gatherer.unknown then
		for item,info in pairs(db.gatherer.unknown) do
			if L.gatherer.types[info.gathertype][item] then
				local newid=L.gatherer.types[info.gathertype][item].id
				mod.core:Print("DEV: moving unknown Gather Item #"..info.id.." > #"..newid.." = "..item)
				for zone_nr,nodelist in pairs(db.gatherer.node) do
					local list=nodelist[info.gathertype]
					if list then 
						for n=1,list.n,1 do
							if list[n].item[info.id] then
								list[n].item[newid]=list[n].item[info.id]
								list[n].item[info.id]=nil
								mod.core:Print("found at "..mod.locator:GetZoneName(zone_nr).." "..list[n].x.."/"..list[n].y)
							end
						end
					end
				end
				db.gatherer.unknown[item]=nil
			else
				L.gatherer.types[info.gathertype][item]=info
			end
		end
	end
	
	for gathertype,list in pairs(L.gatherer.types) do
		for item,info in pairs(list) do
			id2item[info.id]={item=item,icon=info.icon,gathertype=gathertype,crate=info.crate}
		end
	end
end


-- -----------------------------------------------------------------
-- CONFIG INITIALIZE
-- -----------------------------------------------------------------
function Me:ConfigInitialize()
	local pfx=mypfx.."_Config"
	widget.config={	mmsize=mod.config:InitializeSlider(pfx.."_MMSize_Set",5,25,1,1,L.gatherer.config.mmsize.text,L.gatherer.config.mmsize.tt),
					wmsize=mod.config:InitializeSlider(pfx.."_WMSize_Set",5,25,1,1,L.gatherer.config.wmsize.text,L.gatherer.config.wmsize.tt),
					mindist=mod.config:InitializeSlider(pfx.."_MinDist_Set",0,200,1,1,L.gatherer.config.mindist.text,L.gatherer.config.mindist.tt),
					maxdist=mod.config:InitializeSlider(pfx.."_MaxDist_Set",0,200,1,1,L.gatherer.config.maxdist.text,L.gatherer.config.maxdist.tt),
					check={},
					unknown={
						title=getglobal(pfx.."_Unknown_Title"),
						up=getglobal(pfx.."_Unknown_Up"),down=getglobal(pfx.."_Unknown_Down"),
						del={[1]=getglobal(pfx.."_Unknown_Del_1"),[2]=getglobal(pfx.."_Unknown_Del_2"),[3]=getglobal(pfx.."_Unknown_Del_3")},
						text={[1]=getglobal(pfx.."_Unknown_Text_1"),[2]=getglobal(pfx.."_Unknown_Text_2"),[3]=getglobal(pfx.."_Unknown_Text_3")},
					},
				  }
	widget.config.mmsize.slider.exitmodule=Me
	widget.config.wmsize.slider.exitmodule=Me
	widget.config.mindist.slider.exitmodule=Me
	widget.config.maxdist.slider.exitmodule=Me
	for typname,typ in pairs(TYPE) do 
		widget.config.check[typ]=getglobal(pfx.."_Show_"..typ)
		getglobal(pfx.."_Show_"..typ.."Text"):SetText(L.gatherer.config[typ])
	end
	mod.config:Register(Me,L.button.gatherer.config,"vZifnabs_Gatherer_Config")
	Me:ResizeIcons()
end

-- -----------------------------------------------------------------
-- CONFIGURATION START
-- -----------------------------------------------------------------
function Me:ConfigStart()
	widget.config.mmsize.slider:SetValue(db.gatherer.mmsize)
	widget.config.wmsize.slider:SetValue(db.gatherer.wmsize)
	widget.config.mindist.slider:SetValue(db.gatherer.mindistance)
	widget.config.maxdist.slider:SetValue(db.gatherer.maxdistance)
	for typname,typ in pairs(TYPE) do 
		widget.config.check[typ]:SetChecked(db.gatherer.show[playername][typ]) 
	end
	config.unknown={top=1,high=0,list={}}
	for item,info in pairs(db.gatherer.unknown) do
		config.unknown.high=config.unknown.high+1
		config.unknown.list[config.unknown.high]={item=item,id=info.id,gathertype=info.gathertype}
	end
	Me:ShowUnknown()
	mod.config:SetUnsavedChanges(Me:ConfigChanged()) 
end

-- -----------------------------------------------------------------
-- CONFIGURATION SAVE
-- -----------------------------------------------------------------
function Me:ConfigSave()
	db.gatherer.mmsize=widget.config.mmsize.slider:GetValue()
	db.gatherer.wmsize=widget.config.wmsize.slider:GetValue()
	db.gatherer.mindistance=widget.config.mindist.slider:GetValue()
	db.gatherer.maxdistance=widget.config.maxdist.slider:GetValue()
	for typname,typ in pairs(TYPE) do
		db.gatherer.show[playername][typ]=(widget.config.check[typ]:GetChecked()~=nil)
	end
	Me:ResizeIcons()
	current_zone=-1 -- just to force recreation of minimap list
	for u=1,config.unknown.high,1 do
		if config.unknown.list[u].delete then
			local delid=config.unknown.list[u].id
			local deltype=config.unknown.list[u].gathertype
			mod.core:Print("deleting "..config.unknown.list[u].item.." type "..deltype.." id "..delid)
			db.gatherer.unknown[config.unknown.list[u].item]=nil
			for zone_nr,nodes in pairs(db.gatherer.node) do
				if nodes[deltype] then
					for n=1,nodes[deltype].n,1 do
						if nodes[deltype][n].item[delid] then
							nodes[deltype][n].item[delid]=nil
							local remain=0
							for id,count in pairs(nodes[deltype][n].item) do remain=remain+1 end
							if remain==0 then 
								nodes[deltype][n]=nodes[deltype][nodes[deltype].n]
								nodes[deltype][nodes[deltype].n]=nil 
								nodes[deltype].n=nodes[deltype].n-1
							end
						end
					end
				end
			end
		end
	end
	Me:RedefineItems()
	current_zone=-1 -- to recreate itemlist
	Me:ConfigStart()
end

-- -----------------------------------------------------------------
-- CONFIGURATION DROP & STOP
-- -----------------------------------------------------------------
function Me:ConfigDrop() Me:ConfigStart() end
function Me:ConfigStop() end

-- -----------------------------------------------------------------
-- SLIDER VALUE CHANGED
-- -----------------------------------------------------------------
function Me:SliderValueChanged(sliderwidget,effectivevalue)
	mod.config:SetUnsavedChanges(Me:ConfigChanged()) 
	if widget.config.wmsize then Me:ResizeIcons(widget.config.mmsize.slider:GetValue(),widget.config.wmsize.slider:GetValue()) end
end

-- -----------------------------------------------------------------
-- CHECKBOX CLICKED
-- -----------------------------------------------------------------
function Me:Check(checkwidget)
	mod.config:SetUnsavedChanges(Me:ConfigChanged())
end

-- -----------------------------------------------------------------
-- BUTTON
-- -----------------------------------------------------------------
function Me:Button(button)
	local parse=mod.core:Split(button:GetName(),"_")
	local what=parse[getn(parse)]
	if what=="Up" then config.unknown.top=math.max(1,config.unknown.top-1)
	elseif what=="Down" then config.unknown.top=math.max(1,math.min(config.unknown.high-2,config.unknown.top+1))
	elseif parse[getn(parse)-1]=="Del" then
		local ux=tonumber(what)+config.unknown.top-1
		config.unknown.list[ux].delete=true
	end
	mod.config:SetUnsavedChanges(Me:ConfigChanged()) 
	Me:ShowUnknown()
end

-- -----------------------------------------------------------------
-- SHOW UNKNOWN
-- -----------------------------------------------------------------
function Me:ShowUnknown()
	for u=1,3,1 do
		local ux=u+config.unknown.top-1
		widget.config.unknown.text[u]:SetText(config.unknown.list[ux].item or "")
		if ux<=config.unknown.high then widget.config.unknown.del[u]:Show() else widget.config.unknown.del[u]:Hide() end
		if not config.unknown.list[ux].delete then widget.config.unknown.del[u]:Enable() else widget.config.unknown.del[u]:Disable() end
	end
end

-- -----------------------------------------------------------------
-- CONFIG CHANGED ?
-- -----------------------------------------------------------------
function Me:ConfigChanged()
	local changed=false
	if widget.config.mmsize.slider:GetValue()~=db.gatherer.mmsize then changed=true end
	if widget.config.wmsize.slider:GetValue()~=db.gatherer.wmsize then changed=true end
	
	if widget.config.mindist.slider:GetValue()~=db.gatherer.mindistance then changed=true end
	if widget.config.maxdist.slider:GetValue()~=db.gatherer.maxdistance then changed=true end
	for typname,typ in pairs(TYPE) do
		local check=(widget.config.check[typ]:GetChecked()~=nil)
		if check~=db.gatherer.show[playername][typ] then changed=true end
	end
	if config.unknown then
		for u=1,config.unknown.high,1 do
			if config.unknown.list[u].delete then changed=true end
		end
	end
	return changed
end

-- -----------------------------------------------------------------
-- RESIZE ICONS
-- -----------------------------------------------------------------
function Me:ResizeIcons(mm,wm) 
	if not mm then mm=db.gatherer.mmsize end
	if not wm then wm=db.gatherer.wmsize end
	for i=1,MAXMM,1 do 
		widget.mm[i]:SetWidth(mm) 
		widget.mm[i]:SetHeight(mm) 
	end
	for i=1,MAXWM,1 do 
		widget.wm[i]:SetWidth(wm) 
		widget.wm[i]:SetHeight(wm) 
	end
end