local Me=CreateFrame("Frame","vZifnabsLocator")
vZifnabs.mod.locator=Me
local mod=vZifnabs.mod
local db=nil
local L={}

local widget={pfx="vZifnabs_Locator"}
local ticktack=0
local zone_name={}
local config={}

local FRACT=1000
local POS_MULTI=100*FRACT
local ZONE_MULTI=POS_MULTI*POS_MULTI

local MMDEFAULTSCALE={X=6.2,Y=4.8}
local ZOOMSCALE={[0]=1,[1]=1.1,[2]=1.3,[3]=1.7,[4]=2.3,[5]=3.3}

-- -----------------------------------------------------------------
-- INITIALIZE
-- -----------------------------------------------------------------
function Me:Initialize()
	db,L=mod.core:InitStart("locator")
		
	mod.core:RegisterEvent("TICK","locator",Me.Tick)
	mod.core:RegisterEvent("WORLD_MAP_UPDATE","locator",Me.WORLD_MAP_UPDATE);
	if not db.locator then db.locator={waypoint={}} end

	-- load zone names by numbers for quicker reference
	for name,entry in pairs(L.locator.zones) do
		zone_name[entry.nr]=name
	end
	
	-- Check for not permanently defined zones
	if db.locator.unknownzones then
		local udef=0
		for name,entry in pairs(db.locator.unknownzones) do
			if name~="next_nr" then
				zone_name[entry.nr]=name
				local count=0
				for name,info in pairs(db.targets.seen) do
					if info.pos then
						for nr,pos in pairs(info.pos) do
							local x,y,zone,zonenr=Me:ParseLoc(pos)
							if zonenr==entry.nr then count=count+1 end
						end
					end
				end
				if count>0 then 
					mod.core:Print(mod.core:Repl(L.msg.undefzonelist,{"<name>","<nr>","<count>"},{name,entry_nr,count}))
					udef=udef+1
				else
					mod.core:Print(mod.core:Repl(L.msg.undefzonedel,{"<name>","<nr>"},{name,entry_nr}))
					db.locator.unknownzones[name]=nil
				end
			end
		end
		if udef==0 then db.locator.unknownzones=nil end
	end
	
	-- Check for not permanently defined mm scales
	if db.locator.mmscale then
		local lz=L.locator.zones
		for zone,scale in pairs(db.locator.mmscale) do
			local dbscale="(x="..(scale.x or MMDEFAULTSCALE.X)..", y="..(scale.y or MMDEFAULTSCALE.Y)..")"
			local lzscale="(x="..MMDEFAULTSCALE.X..", y="..MMDEFAULTSCALE.Y..")"
			if lz[zone] then lzscale="(x="..(lz[zone].x or MMDEFAULTSCALE.X)..", y="..(lz[zone].y or MMDEFAULTSCALE.Y)..")" end
			if not lz[zone] then 
				mod.core:Print("INTERNAL ERROR: db mmscale overwrite "..dbscale.." for unexistant zone "..zone)
			elseif dbscale==lzscale then
				db.locator.mmscale[zone]=nil
			else
				mod.core:Print(mod.core:Repl(L.msg.scaleovw,{"<dbscale>","<zone>","<lzscale>"},{dbscale,zone,lzscale}))
			end
		end
	end
	
	if db.locator.waypoint[UnitName("player")] and not db.locator.waypoint[UnitName("player")].pos then db.locator.waypoint[UnitName("player")]=nil end
	widget.postext=getglobal(widget.pfx.."Pos_Text")
	widget.postext:SetTextColor(1,1,1,1)

	widget.mm=getglobal(widget.pfx.."_MM")
	widget.mm:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", edgeFile = nil});
	widget.mm:SetBackdropColor(0,0,1,1);
	widget.wm={}
	for nr=1,10,1 do
		widget.wm[nr]=getglobal(widget.pfx.."_WM_"..nr)
		widget.wm[nr]:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", edgeFile = nil});
		widget.wm[nr]:SetBackdropColor(1,0,0,1);
		widget.wm[nr]:Hide()
	end
	widget.wminfo=getglobal(widget.pfx.."_WMinfo")
	widget.wmtext=getglobal(widget.pfx.."_WMinfo_Text")
	widget.wmtext:SetTextColor(1,0,0,1)
	
	Me:ConfigInitialize()
	mod.core:InitStop("locator")
end

-- -----------------------------------------------------------------
-- BUTTON
-- -----------------------------------------------------------------
function Me:Button(button,fnk)
	local x,y,zone,pos=Me:GetLoc()
	if fnk=="MM" and db.locator.waypoint[UnitName("player")] then	
		db.locator.waypoint[UnitName("player")]=nil
		for n=1,10,1 do
			widget.wm[n]:Hide()
		end
		widget.wminfo:Hide()
	elseif fnk=="MM" then
		Me:SetMM(zone.." "..date("%d.%m.%y, %H:%M:%S",time()),{["n"]=1,[1]=pos})
	elseif fnk=="WM" then
		local split=mod.core:Split(button:GetName(),"_")
		local n=tonumber(split[4])
		if db.locator.waypoint[UnitName("player")].pos[n] then db.locator.waypoint[UnitName("player")].pos.n=n end
	end
end

-- -----------------------------------------------------------------
-- SET MM
-- -----------------------------------------------------------------
function Me:SetMM(setname,pos)
	local waypoint={name=setname,pos={}}
	waypoint.pos.n=pos.n
	for n=1,10,1 do
		waypoint.pos[n]=pos[n]
	end
	db.locator.waypoint[UnitName("player")]=waypoint
	
end

-- -----------------------------------------------------------------
-- WORLD MAP
-- -----------------------------------------------------------------
function Me:WORLD_MAP_UPDATE(a1,a2,a3,a4)
	local waypoint=db.locator.waypoint[UnitName("player")]
	if waypoint and WorldMapFrame:IsVisible() then 
		local continent = GetCurrentMapContinent()
		local zonelist={GetMapZones(continent)}
		local zone=zonelist[GetCurrentMapZone()]
		if L.locator.zonereplace[zone] then zone=L.locator.zonereplace[zone] end

		widget.wmtext:SetText("")
		for n=1,10,1 do
			local pos=waypoint.pos[n]
			local show=pos
			if pos then 
				local wx,wy,wzone=Me:ParseLoc(pos)
				wx=wx*WorldMapDetailFrame:GetWidth()/100
				wy=wy*WorldMapDetailFrame:GetHeight()/100
				widget.wm[n]:SetPoint("CENTER","WorldMapDetailFrame","TOPLEFT",wx,-wy)
				show=(wzone==zone)
			end
			if show then widget.wm[n]:Show() else widget.wm[n]:Hide() end
		end
	end
end

-- -----------------------------------------------------------------
-- CLOCK
-- -----------------------------------------------------------------
function Me:Tick()
	local x,y,zone=Me:GetLoc()
	local waypoint=db.locator.waypoint[UnitName("player")]
	local color={red=0,green=0,blue=1}
	local mmradius=Minimap:GetWidth() / 2
	local point={x=0,y=-(mmradius-8),zx=0,zy=0,tt=""}
	local tick=0
	if waypoint then 
		color={red=1,green=1,blue=1}
		local ptr=waypoint.pos.n
		for n=1,10,1 do
			local pos=waypoint.pos[ptr]
			if pos then
				local wx,wy,wzone=Me:ParseLoc(pos)
				if wzone==zone and not point.pos then point.pos=pos end
			end
			ptr=ptr-1
			if ptr==0 then ptr=10 end
		end
		if not point.pos then point.pos=waypoint.pos[waypoint.pos.n] end
		local wx,wy,wzone=Me:ParseLoc(point.pos)

		if wzone~=zone then point.tt=wzone.." ("..wx.."/"..wy..")"
		else
			local distance=math.sqrt((x-wx)*(x-wx)+(y-wy)*(y-wy))
			point.tt="("..wx.."/"..wy..") >>"..math.floor(10*distance)/10
			local unscaledx,unscaledy,mmscalex,mmscaley=Me:GetMinimapScale()
			local mmx=(wx-x)*mmradius/mmscalex
			local mmy=(y-wy)*mmradius/mmscaley
			local mmdistance=sqrt(mmx * mmx + mmy * mmy)
			if mmdistance < mmradius-4 then
				point.x=mmx
				point.y=mmy
			else
				point.x=mmx*(mmradius-4)/mmdistance
				point.y=mmy*(mmradius-4)/mmdistance
			end	
			if mmdistance<mmradius then			color={red=0,green=1,blue=0}
			elseif mmdistance<2*mmradius then	color={red=0.5,green=1,blue=0}
			elseif mmdistance<3*mmradius then	color={red=1,green=0.5,blue=0}
			end
			ticktack=ticktack+1
			if ticktack>3 then ticktack=0 end
			tick=math.floor(ticktack/2)
		end
	end
	if (waypoint) then point.tt=waypoint.name.."\n"..point.tt end
	widget.mm:SetPoint("CENTER","Minimap","CENTER",point.x,point.y)
	if (tick==0) then 
		widget.mm:SetBackdropColor(color.red,color.green,color.blue,1)
	else 
		widget.mm:SetBackdropColor(1,1,1,1)
	end
	widget.mm.tooltipText=point.tt
	
	x=math.floor((100*x+5)/10)/10
	y=math.floor((100*y+5)/10)/10
	if not string.find(" "..x,"[.]") then x=x..".0" end
	if not string.find(" "..y,"[.]") then y=y..".0" end
	widget.postext:SetText(x.." / "..y)	
end

-- -----------------------------------------------------------------
-- GET LOCATION
-- -----------------------------------------------------------------
function Me:GetLoc(zone, x, y)
	if not zone then 
		zone=Me:GetZoneName()
		x, y = GetPlayerMapPosition("player")
	end
	local zone_nr=Me:GetZoneNr(zone)
	x=math.min(POS_MULTI-1,math.floor(POS_MULTI*x+0.5))
	y=math.min(POS_MULTI-1,math.floor(POS_MULTI*y+0.5))
	return x/FRACT,y/FRACT,zone,ZONE_MULTI*(zone_nr or 0)+POS_MULTI*x+y,zone_nr
end

-- -----------------------------------------------------------------
-- PARSE LOCATION NUMBER
-- -----------------------------------------------------------------
function Me:ParseLoc(loc)
	local zone_nr=math.floor(loc/ZONE_MULTI)
	loc=loc-zone_nr*ZONE_MULTI
	local x=math.floor(loc/POS_MULTI)
	local y=loc-x*POS_MULTI
	return x/FRACT,y/FRACT,zone_name[zone_nr],zone_nr
end

-- -----------------------------------------------------------------
-- GET ZONE NUMBER
-- -----------------------------------------------------------------
function Me:GetZoneNr(zone)
	local zone_nr=nil
	if L.locator.zones[zone] then 
		zone_nr=L.locator.zones[zone].nr
	elseif db.locator.unknownzones and db.locator.unknownzones[zone] then
		zone_nr=db.locator.unknownzones[zone].nr
	elseif zone and zone~="" then
		if not db.locator.unknownzones then db.locator.unknownzones={next_nr=999} end
		zone_nr=db.locator.unknownzones.next_nr
		db.locator.unknownzones.next_nr=db.locator.unknownzones.next_nr-1
		db.locator.unknownzones[zone]={nr=zone_nr,addby=UnitName("player"),time=time()}
	end
	return zone_nr
end

-- -----------------------------------------------------------------
-- GET ZONE NUMBER
-- -----------------------------------------------------------------
function Me:GetZoneName(continent,zonenr)
	local here=(not zonenr)
	if here then
		SetMapToCurrentZone()
		zonenr=GetCurrentMapZone()
		continent=GetCurrentMapContinent()
	end
	local zones={GetMapZones(continent)}
	local zone=zones[zonenr]
	if here and not zone then zone=GetZoneText() end
	return zone
end

-- -----------------------------------------------------------------
-- GET MINIMAP SCALE
-- -----------------------------------------------------------------
function Me:GetMinimapScale(zone)
	if not zone then zone=Me:GetZoneName() end
	local scalex=MMDEFAULTSCALE.X
	local scaley=MMDEFAULTSCALE.Y
	if L.locator.zones[zone] then
		scalex=L.locator.zones[zone].x or scalex
		scaley=L.locator.zones[zone].y or scaley
	end
	if db.locator.mmscale and db.locator.mmscale[zone] then
		scalex=db.locator.mmscale[zone].x or scalex
		scaley=db.locator.mmscale[zone].y or scaley
	end
	if config.scalezone==zone then
		scalex=config.val.scalex.now or scalex
		scaley=config.val.scaley.now or scaley
	end
	local zoom=Minimap:GetZoom()
	return scalex,scaley,scalex/ZOOMSCALE[zoom],scaley/ZOOMSCALE[zoom]
end

-- -----------------------------------------------------------------
-- CONFIGURATION INITIALIZE
-- -----------------------------------------------------------------
function Me:ConfigInitialize()
	local pfx=widget.pfx.."_Config"

	widget.mmscalex=mod.config:InitializeSlider(pfx.."_MMScale_X",1,400,1,10,L.locator.mmscale.x.text,L.locator.mmscale.x.tt)
	widget.mmscalex.slider.exitmodule=Me
	widget.mmscalex.slider.valuename="scalex"
	widget.mmscaley=mod.config:InitializeSlider(pfx.."_MMScale_Y",1,400,1,10,L.locator.mmscale.y.text,L.locator.mmscale.y.tt)
	widget.mmscaley.slider.exitmodule=Me
	widget.mmscaley.slider.valuename="scaley"
	getglobal(pfx.."_MMScale_Text"):SetText(L.locator.mmscale.text)
	widget.mmzone=getglobal(pfx.."_MMScale_Zone")
	
	mod.config:Register(Me,L.button.locator.config,"vZifnabs_Locator_Config")
end

-- -----------------------------------------------------------------
-- CONFIGURATION START
-- -----------------------------------------------------------------
function Me:ConfigStart()
	local scalex,scaley=Me:GetMinimapScale(config.scalezone)
	config={prepare=true,scalezone=Me:GetZoneName(),val={
		scalex={init=scalex,now=scalex},
		scaley={init=scaley,now=scaley},
	}}
	widget.mmzone:SetText(config.scalezone)
	widget.mmscalex.slider:SetValue(10*scalex)
	widget.mmscaley.slider:SetValue(10*scaley)
	config.prepare=false
end

-- -----------------------------------------------------------------
-- CONFIGURATION SAVE
-- -----------------------------------------------------------------
function Me:ConfigSave()
	local mmzone=config.scalezone
	db.locator.mmscale[mmzone]={x=config.val.scalex.now,y=config.val.scaley.now}
	mod.core:Print(mod.core:Repl(L.msg.mmscale,{"<zone>","<x>","<y>"},{mmzone,db.locator.mmscale[mmzone].x,db.locator.mmscale[mmzone].y}))
end

-- -----------------------------------------------------------------
-- CONFIGURATION DROP & STOP
-- -----------------------------------------------------------------
function Me:ConfigDrop() Me:ConfigStart() end
function Me:ConfigStop() config={} end

-- -----------------------------------------------------------------
-- SLIDER VALUE CHANGED
-- -----------------------------------------------------------------
function Me:SliderValueChanged(widget,effectivevalue)
	config.val[widget.valuename].now=effectivevalue
	local changes=false
	for name,data in pairs(config.val) do changes=changes or data.init~=data.now end
	mod.config:SetUnsavedChanges(changes) 
end